/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  Changed by Andrew Crew Kuznetsov <andrewcrew@rambler.ru>
 */

#include "xProgram.h"
#include "conv.h"
#include "detection.h"
#include "error.h"

//global
static status exit_flag = 1;

//set exit flag
void xProgram_exit ()
{
	exit_flag = 0;
}

//test exit flag
bool xProgram_isExiting ()
{
	if (exit_flag)
		return FALSE;
	else
		return TRUE;
}

//allocate structs and connect to X
xProgramInfo *xProgram_init (xconf* xnc)
{
	xProgramInfo *toInit = NULL;

	//allocate & init huge xProgramInfo struct
	toInit = (xProgramInfo *) malloc (sizeof (xProgramInfo));
	if (!toInit) {
		send_error (ERROR, "Bad alloc in xProgramInfo\n", NULL);
		return NULL;
	}

	toInit->xobject = xObject_init ();	//program window

	//connect to X Server
	if (ERROR == xObject_connect (toInit->xobject)) {
		return NULL;
	}
	//set func to catch xlib Errors
	xObject_setXErrFunc (handleXErrors);

	toInit->xword = xString_init (MAX_WORD_LENGTH);	//current word flush buffer
	toInit->xstring = xString_init (MAX_STRING_LENGTH);	//current string buffer
	toInit->xselection = xSelection_init (xObject_getDisplay (toInit->xobject));	//X selection processor
	toInit->xswitchlang = xSwitchLang_init (xObject_getDisplay (toInit->xobject));	//KBD-based layout switch
	toInit->xevent = xEvent_init ();	//X event processor
	toInit->xifocus = xIFocus_init ();	//X Input Focus and Pointer processor
	toInit->xklbuffer = xKeyLogBuffer_init ();	//Input word buffer
	toInit->xbtable = xBindingTable_init (DEFAULT_BIND_TABLE_SIZE);	//Key Bind Table
	if (!(toInit->xobject) || !(toInit->xstring) || !(toInit->xselection)
	    || !(toInit->xevent) || !(toInit->xifocus)
	    || !(toInit->xswitchlang) || !(toInit->xklbuffer)
	    || !(toInit->xbtable) || !(toInit->xword)) {
		xObject_free (toInit->xobject);
		xString_free (toInit->xstring);
		xSelection_free (toInit->xselection);
		xEvent_free (toInit->xevent);
		xIFocus_free (toInit->xifocus);
		xSwitchLang_free (toInit->xswitchlang);
		xKeyLogBuffer_free (toInit->xklbuffer);
		xBindingTable_free (toInit->xbtable);

		free (toInit);

		send_error (ERROR, "Bad alloc in xProgramInfo\n", NULL);
		return NULL;
	}

	//create program window
	if (ERROR == xObject_createWindow (toInit->xobject))
		return NULL;
	//selecting no core event mask. Remember: selection notify event have no mask.
	//KeyPressMask|ButtonReleaseMask|ButtonPressMask|StructureNotifyMask
	xObject_setEventMask (toInit->xobject, None);
	
	//xstring & xword xobject init
	toInit->xstring->home->xDisplay = xObject_getDisplay (toInit->xobject);
	toInit->xstring->home->xWindow = xObject_getWindow (toInit->xobject);
	toInit->xword->home->xDisplay = xObject_getDisplay (toInit->xobject);
	toInit->xword->home->xWindow = xObject_getWindow (toInit->xobject);
	//xselection
	toInit->xselection->requestor->xDisplay = xObject_getDisplay (toInit->xobject);
	toInit->xselection->requestor->xWindow = xObject_getWindow (toInit->xobject);
	toInit->xselection->owner->xDisplay = toInit->xobject->xDisplay;
	//xevent
	xEvent_setCommunicator (toInit->xevent, toInit->xobject);
	//xifocus
	toInit->xifocus->owner->xDisplay = toInit->xobject->xDisplay;

	//set bindings
	int i;
	for(i = 1; i<DEFAULT_BIND_TABLE_SIZE; i++)
		xKeyLogManager_bind (toInit->xbtable, xnc->btable[i].key,
						xnc->btable[i].key_modifier, i);
	
	XSynchronize (toInit->xobject->xDisplay, TRUE);
	XFlush (toInit->xobject->xDisplay);
	return toInit;
}

//close connection & free xProgramInfo struct
status xProgram_free (xProgramInfo * toDelete)
{
	status err = SUCCESS;

	XDestroyWindow (xObject_getDisplay (toDelete->xobject),
			xObject_getWindow (toDelete->xobject));
	XCloseDisplay (xObject_getDisplay (toDelete->xobject));

	if (toDelete) {
		if ((xObject_free (toDelete->xobject) != SUCCESS) ||
		    (xString_free (toDelete->xstring) != SUCCESS) ||
		    (xString_free (toDelete->xword) != SUCCESS) ||
		    (xSelection_free (toDelete->xselection) != SUCCESS) ||
		    (xEvent_free (toDelete->xevent) != SUCCESS) ||
		    (xIFocus_free (toDelete->xifocus) != SUCCESS) ||
		    (xSwitchLang_free (toDelete->xswitchlang) != SUCCESS) ||
		    (xKeyLogBuffer_free (toDelete->xklbuffer) != SUCCESS) ||
		    (xBindingTable_free (toDelete->xbtable) != SUCCESS))
			err = WARNING;
		free (toDelete);
	}
	else
		err = WARNING;

	return err;
}

//get info about current input focus owner
bool xProgram_update (xProgramInfo * xinfo, app_list * sup_apps_list)
{
	status err = SUCCESS;

	//setting xEvent reciever params to updated focus window params
	err = xClientManager_getClientInfo (xinfo->xevent, xinfo->xifocus);

	//if focus owner window is set to None (ERROR) 
	if (ERROR == err) {
		return FALSE;
	}

	if (!xClientManager_testAppName (xinfo->xifocus, sup_apps_list))
			return FALSE;

	//begin listening application
	xClientManager_beginToListenClient (xinfo->xifocus);
	
	return TRUE;
}

//log keys
num xProgram_processInput (xProgramInfo * xinfo)
{
	char *res_str = NULL;
	num res_chars = 0;

	//return early if bind detected
	res_str =
		xKeyLogManager_getStringFromEventQueue (xinfo->xklbuffer,
							xinfo->xevent, xinfo->xifocus, xinfo->xword,
							xinfo->xstring, xinfo->xswitchlang,
							xinfo->xbtable);
	if (res_str)
		res_chars = strlen (res_str);
	else
		res_chars = 0;

	return res_chars;
}

//process bind action
bool xProgram_performCurrentBind (xProgramInfo * xinfo, xconf* xnc)
{
	bool back_to_input;
	status action = xBindingTable_getCurrentAction (xinfo->xbtable);
	status lang, err;
	char sym;
	char *low_word = NULL;
	if (action != BA_NONE) {
		switch (xnc->CurrentMode) {
				//default auto mode
			case AUTO_BY_KEYLOG:
				switch (xBindingTable_getCurrentAction (xinfo->xbtable)) {
						//user need to cancel last change
					case BA_CHANGE_WORD:
						lang = xProgram_getCurrentLang (xinfo);
						if (ENGLISH == lang) {
							lang = RUSSIAN;
							send_error (DEBUG, "Back -> to Rus\n",
								    NULL);
						}
						else {
							lang = ENGLISH;
							send_error (DEBUG, "Back -> to Eng\n",
								    NULL);
						}

						xKeyLogManager_flushBuffer (xinfo->xklbuffer,
									    xinfo->xword);
						err = xProgram_changeWordBySelection (xinfo, lang);
						back_to_input = TRUE;
						// Buffer clearing
						sym = xEvent_getCurrentChar (xinfo->xevent);
						xString_catContent (xinfo->xklbuffer->buffer, xinfo->xstring);
						xKeyLogBuffer_clear (xinfo->xklbuffer);
						send_error (DEBUG, "KLB_FLUSH\n", NULL);
						send_error (DEBUG, "Current string=%s\n", &xinfo->xstring->content);
						free (low_word);
						break;
					case BA_CHANGE_MODE:
						xnc->CurrentMode = MANUAL_ON_SHORTCUT;
						send_error(DEBUG, "Mode -> to Manual\n", NULL);
						back_to_input = FALSE;
						break;
					case BA_CHANGE_SELECT:
						send_error(DEBUG, "Change Selected text\n", NULL);
						err = xEvent_getSelectedText (xinfo->xevent);
						back_to_input = FALSE;
						break;
					default:
						break;
				}
				break;
				//default manual mode
			case MANUAL_ON_SHORTCUT:
				switch (action) {
						//user need to change last word
					case BA_CHANGE_WORD:
						lang = xProgram_getCurrentLang (xinfo);
						if (ENGLISH == lang) {
							lang = RUSSIAN;
							send_error (DEBUG, "Eng -> to Rus\n", NULL);
						}
						else {
							lang = ENGLISH;
							send_error (DEBUG, "Rus -> to Eng\n", NULL);
						}
						send_error (DEBUG,
							    "Processing word: %s\n",
							    &(xinfo->xword->content));
						xKeyLogManager_flushBuffer (xinfo->xklbuffer,
									    xinfo->xword);
						err = xProgram_changeWordBySelection (xinfo, lang);
						back_to_input = TRUE;
						// Buffer clearing
						sym = xEvent_getCurrentChar (xinfo->xevent);
						xString_catContent (xinfo->xklbuffer->buffer, xinfo->xstring);
						xKeyLogBuffer_clear (xinfo->xklbuffer);
						send_error (DEBUG, "KLB_FLUSH\n", NULL);
						send_error (DEBUG, "Current string=%s\n", &xinfo->xstring->content);
						free (low_word);
						break;
						//user need to change current phrase
					case BA_CHANGE_STRING:
						lang = xProgram_getCurrentLang (xinfo);
						if (ENGLISH == lang) {
							lang = RUSSIAN;
							send_error (DEBUG, "Eng -> to Rus\n", NULL);
						}
						else {
							lang = ENGLISH;
							send_error (DEBUG, "Rus -> to Eng\n", NULL);
						}
						xKeyLogManager_flushBuffer (xinfo->xklbuffer,
									    xinfo->xword);
						err = xProgram_changeStringBySelection (xinfo,
											lang);
						back_to_input = TRUE;
						// Buffer clearing
						sym = xEvent_getCurrentChar (xinfo->xevent);
						xString_catContent (xinfo->xklbuffer->buffer, xinfo->xstring);
						xKeyLogBuffer_clear (xinfo->xklbuffer);
						send_error (DEBUG, "KLB_FLUSH\n", NULL);
						send_error (DEBUG, "Current string=%s\n", &xinfo->xstring->content);
						free (low_word);
						break;
					case BA_CHANGE_MODE:
						xnc->CurrentMode = AUTO_BY_KEYLOG;
						send_error(DEBUG, "Mode -> to Auto\n", NULL);
						back_to_input = FALSE;
						break;
					case BA_CHANGE_SELECT:
						send_error(DEBUG, "Change Selected text\n", NULL);
						back_to_input = FALSE;
						break;
					default:
						break;
				}
				break;
			default:
				break;
		}
		//notify we'll have no bind
		xBindingTable_setCurrentAction (xinfo->xbtable, BA_NONE);
	}
	else
		//no bind
		back_to_input = FALSE;

	return back_to_input;
}

//perform mode actions
status xProgram_performCurrentMode (xProgramInfo * xinfo, xconf * xnc)
{
	num cnp = 0;
	char *logged_word = xinfo->xword->content;
	size word_len = xString_getLength (xinfo->xword);
	char *low_word = NULL;
	//char *ret_word = NULL;
	status lang, err = SUCCESS;
	//num itr;
	char sym;
	int wordlen = 0;
	char *ru = xnc->rudict;
	char *en = xnc->endict;
	char word[66];
	char last_sym[]=" ,.?/!@#$%^&*()_+-;:\"";
	char last_symbol;
	int i = -1;
	int j;

	switch (xnc->CurrentMode) {
			//default auto mode
		case AUTO_BY_KEYLOG:
			//get current layout
			lang = xProgram_getCurrentLang (xinfo);
			send_error (DEBUG, "Processing word: '%s'\n", &logged_word);
			//test word by lang algorythm
			low_word = AllToLower ((const char *) logged_word);
			if (low_word == NULL)
				return err;
			// using old algorithm language detect
			if (xnc->LangDetectMode==0) {
				send_error (DEBUG, "0 - Is Use Old Algorithm (language %d)\n", &lang);
				cnp = isCyrWord ((const char *) low_word);
				if ((cnp > xnc->ConvPoint) && (ENGLISH == lang)) {
					//process string in client window
					err = xProgram_changeWordBySelection (xinfo, RUSSIAN);
					send_error (DEBUG, "Eng -> to Rus\n", NULL);
				}
				else if ((cnp < xnc->ConvPoint) && (RUSSIAN == lang)) {
					//process string in client window
					err = xProgram_changeWordBySelection (xinfo, ENGLISH);
					send_error (DEBUG, "Rus -> to Eng\n", NULL);
				}
				else
					send_error (DEBUG, "Nothing to change\n", NULL);
			}
			// using new algorithm language detect	
			else if (xnc->LangDetectMode==1) {
				send_error (DEBUG, "1 - Is Use New Algorithm (language %d)\n", &lang);                             
				// change low_word for dictionary search
				if (strchr(last_sym, low_word[strlen((const char*) low_word)-1]) != NULL)
				{
					wordlen = strlen((const char*) low_word);
					last_symbol = low_word[strlen((const char*) low_word)-1];
					low_word[wordlen-1] = '\0';		
				}		
				word[++i] = '|';
				for (j = 0; j < strlen((const char*) low_word); j++){
					if (low_word[j] != ' ')
						word[++i] = low_word[j];
				}
				word[++i] = '|';
				word[++i] = '\0';
				//printf("word: '%s'\n", word);
				//printf("dict ru: '%s'\n", ru);
				//printf("dict en: '%s'\n", en);
				// check russian dictionary 
      			if (strstr(ru, word) != NULL) 
					cnp = 1;
				// check english dictionary    
      			if (strstr(en, word) != NULL) 
					cnp = -1;
				// if not find in dictionary, use heuristic detection	
				// restore low_word				
				i = -1;
				low_word = AllToLower ((const char *) logged_word);

				for (j = 0; j < strlen((const char*) low_word); j++){
					if (low_word[j] != ' ')
						low_word[++i] = low_word[j];
				}
				low_word[++i] = '\0';
				//printf("low_word: '%s'\n", low_word);
				// heuristic search
				if (cnp == 0) 
					cnp = WhatLang ((const char *) low_word);
				//printf("low_word: '%s'\n", low_word);				
				
				if ((cnp > 0) && (ENGLISH == lang)) {
					//process string in client window
					err = xProgram_changeWordBySelection (xinfo, RUSSIAN);
					send_error (DEBUG, "Eng -> to Rus\n", NULL);
				}
				else if ((cnp < 0) && (RUSSIAN == lang)) {
					//process string in client window
					err = xProgram_changeWordBySelection (xinfo, ENGLISH);
					send_error (DEBUG, "Rus -> to Eng\n", NULL);
				}
				else
					send_error (DEBUG, "Nothing to change\n", NULL);
				
			}
			else
 				return err;
			// Buffer clearing
			sym = xEvent_getCurrentChar (xinfo->xevent);
			xString_catContent (xinfo->xklbuffer->buffer, xinfo->xstring);
			xKeyLogBuffer_clear (xinfo->xklbuffer);
			send_error (DEBUG, "KLB_FLUSH\n", NULL);
			send_error (DEBUG, "Current string=%s\n", &xinfo->xstring->content);
			//free (low_word);
			break;
		case MANUAL_ON_SHORTCUT:
			break;
		default:
			break;
	}
	return err;
}

//change last word in window
status xProgram_changeWordBySelection (xProgramInfo * xinfo, status lang)
{
	status err = SUCCESS;
	size word_len = 0;
	char *word = NULL;

	if (!xString_getLength (xinfo->xword)) {
		word = (char *) GetLastWord (xinfo->xstring->content);
		send_error (DEBUG, "Get word='%s' from string buffer\n", &word);
		if (!word)
			return WARNING;
		xString_setContent (xinfo->xword, word);
		free (word);
	}

	word_len = xString_getLength (xinfo->xword);
	//delete old word
	err = xSelectManager_deleteNCharsInCursorStrokeToStart (xinfo->xevent, word_len);
	//write changed word
	err = xSelectManager_sendStringToFocus (xinfo->xword,
						xinfo->xswitchlang, xinfo->xevent, lang);

//      xString_setContent(xinfo->xword, NULL);
//      xString_setContent(xinfo->xstring, NULL);

	return err;
}

//change last phrase in window
status xProgram_changeStringBySelection (xProgramInfo * xinfo, status lang)
{
	status err = SUCCESS;
	size str_len;

	if (xString_getLength (xinfo->xstring) == 0)
		return WARNING;
	//cat last word buffer & current phrase buffer
	xString_catContent (xinfo->xword, xinfo->xstring);
	str_len = xString_getLength (xinfo->xstring);

	//delete old phrase
	err = xSelectManager_deleteNCharsInCursorStrokeToStart (xinfo->xevent, str_len);
	//write changed phrase
	err = xSelectManager_sendStringToFocus (xinfo->xstring,
						xinfo->xswitchlang, xinfo->xevent, lang);
	//zero current buffer phrase & word
	xString_setContent (xinfo->xstring, NULL);
	xString_setContent (xinfo->xword, NULL);

	return err;
}

//get current layout
status xProgram_getCurrentLang (xProgramInfo * xinfo)
{
	return (xinfo->xswitchlang->CurrentLang);
}
